"""
Tests for `sourmash prefetch` command-line and API functionality.
"""

import os
import csv
import gzip
import pytest
import glob
import random
from sourmash.search import PrefetchResult

import sourmash_tst_utils as utils
import sourmash
from sourmash_tst_utils import SourmashCommandFailed
from sourmash import SourmashSignature, sourmash_args
from sourmash.signature import (
    save_signatures_to_json,
    load_signatures_from_json,
    load_one_signature_from_json,
)


def approx_eq(val1, val2):
    if val1 == "":
        return val1 == val2
    return round(float(val1), 3) == round(float(val2), 3)


def test_prefetch_basic(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    c.run_sourmash("prefetch", "-k", "31", sig47, sig63, sig2, sig47, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0

    assert (
        "WARNING: no output(s) specified! Nothing will be saved from this prefetch!"
        in c.last_result.err
    )
    assert "selecting specified query k=31" in c.last_result.err
    assert (
        "loaded query: NC_009665.1 Shewanella baltica... (k=31, DNA)"
        in c.last_result.err
    )
    assert (
        "query sketch has scaled=1000; will be dynamically downsampled as needed"
        in c.last_result.err
    )

    err = c.last_result.err
    assert "loaded 5 total signatures from 3 locations." in err
    assert "after selecting signatures compatible with search, 3 remain." in err

    assert "total of 2 matching signatures." in c.last_result.err
    assert (
        "of 5177 distinct query hashes, 5177 were found in matches above threshold."
        in c.last_result.err
    )
    assert "a total of 0 query hashes remain unmatched." in c.last_result.err


def test_prefetch_select_query_ksize(runtmp, linear_gather):
    # test prefetch where query and subject db both have multiple ksizes
    c = runtmp

    ss = utils.get_test_data("GCF_000005845.2_ASM584v2_genomic.fna.gz.sig")

    c.run_sourmash("prefetch", ss, ss, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert (
        "of 4476 distinct query hashes, 4476 were found in matches above threshold."
        in c.last_result.err
    )


def test_prefetch_subject_scaled_is_larger(runtmp, linear_gather):
    # test prefetch where subject scaled is larger
    c = runtmp

    # make a query sketch with scaled=1000
    fa = utils.get_test_data("genome-s10.fa.gz")
    c.run_sourmash("sketch", "dna", fa, "-o", "query.sig")
    assert os.path.exists(runtmp.output("query.sig"))

    # this has a scaled of 10000, from same genome:
    against1 = utils.get_test_data("scaled/genome-s10.fa.gz.sig")
    against2 = utils.get_test_data("scaled/all.sbt.zip")
    against3 = utils.get_test_data("scaled/all.lca.json")

    # run against large scaled, then small (self)
    c.run_sourmash(
        "prefetch",
        "query.sig",
        against1,
        against2,
        against3,
        "query.sig",
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert "total of 8 matching signatures." in c.last_result.err
    assert (
        "of 48 distinct query hashes, 48 were found in matches above threshold."
        in c.last_result.err
    )
    assert (
        "final scaled value (max across query and all matches) is 10000"
        in c.last_result.err
    )


def test_prefetch_subject_scaled_is_larger_outsigs(runtmp, linear_gather):
    # test prefetch where subject scaled is larger -- output sigs
    c = runtmp

    # make a query sketch with scaled=1000
    fa = utils.get_test_data("genome-s10.fa.gz")
    c.run_sourmash("sketch", "dna", fa, "-o", "query.sig")
    assert os.path.exists(runtmp.output("query.sig"))

    # this has a scaled of 10000, from same genome:
    against1 = utils.get_test_data("scaled/genome-s10.fa.gz.sig")
    against2 = utils.get_test_data("scaled/all.sbt.zip")
    against3 = utils.get_test_data("scaled/all.lca.json")

    # run against large scaled, then small (self)
    c.run_sourmash(
        "prefetch",
        "query.sig",
        against1,
        against2,
        against3,
        "query.sig",
        linear_gather,
        "--save-matches",
        "matches.sig",
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert "total of 8 matching signatures." in c.last_result.err
    assert (
        "of 48 distinct query hashes, 48 were found in matches above threshold."
        in c.last_result.err
    )
    assert (
        "final scaled value (max across query and all matches) is 10000"
        in c.last_result.err
    )

    # make sure non-downsampled sketches were saved.
    matches = sourmash.load_file_as_signatures(runtmp.output("matches.sig"))
    scaled_vals = set([match.minhash.scaled for match in matches])
    assert 1000 in scaled_vals
    assert 10000 in scaled_vals
    assert len(scaled_vals) == 2


def test_prefetch_query_abund(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch w/abund query
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("track_abund/47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    c.run_sourmash("prefetch", "-k", "31", sig47, sig63, sig2, sig47, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0

    assert (
        "WARNING: no output(s) specified! Nothing will be saved from this prefetch!"
        in c.last_result.err
    )
    assert "selecting specified query k=31" in c.last_result.err
    assert (
        "loaded query: NC_009665.1 Shewanella baltica... (k=31, DNA)"
        in c.last_result.err
    )
    assert (
        "query sketch has scaled=1000; will be dynamically downsampled as needed"
        in c.last_result.err
    )

    assert "total of 2 matching signatures." in c.last_result.err
    assert (
        "of 5177 distinct query hashes, 5177 were found in matches above threshold."
        in c.last_result.err
    )
    assert "a total of 0 query hashes remain unmatched." in c.last_result.err


def test_prefetch_subj_abund(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch w/abund signature.
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("track_abund/63.fa.sig")

    c.run_sourmash("prefetch", "-k", "31", sig47, sig63, sig2, sig47, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0

    assert (
        "WARNING: no output(s) specified! Nothing will be saved from this prefetch!"
        in c.last_result.err
    )
    assert "selecting specified query k=31" in c.last_result.err
    assert (
        "loaded query: NC_009665.1 Shewanella baltica... (k=31, DNA)"
        in c.last_result.err
    )
    assert (
        "query sketch has scaled=1000; will be dynamically downsampled as needed"
        in c.last_result.err
    )

    assert "total of 2 matching signatures." in c.last_result.err
    assert (
        "of 5177 distinct query hashes, 5177 were found in matches above threshold."
        in c.last_result.err
    )
    assert "a total of 0 query hashes remain unmatched." in c.last_result.err


def test_prefetch_csv_out(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with CSV output
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    csvout = c.output("out.csv")

    c.run_sourmash(
        "prefetch", "-k", "31", sig47, sig63, sig2, sig47, "-o", csvout, linear_gather
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(csvout)

    expected_intersect_bp = [2529000, 5177000]
    with open(csvout, newline="") as fp:
        r = csv.DictReader(fp)
        for row, expected in zip(r, expected_intersect_bp):
            print(row)
            assert int(row["intersect_bp"]) == expected


def test_prefetch_csv_gz_out(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with CSV output to a .gz file
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    csvout = c.output("out.csv.gz")

    c.run_sourmash(
        "prefetch", "-k", "31", sig47, sig63, sig2, sig47, "-o", csvout, linear_gather
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(csvout)

    expected_intersect_bp = [2529000, 5177000]
    with gzip.open(csvout, "rt", newline="") as fp:
        r = csv.DictReader(fp)
        for row, expected in zip(r, expected_intersect_bp):
            print(row)
            assert int(row["intersect_bp"]) == expected


def test_prefetch_matches(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with --save-matches
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    matches_out = c.output("matches.sig")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "--save-matches",
        matches_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(matches_out)

    sigs = sourmash.load_file_as_index(matches_out)

    expected_matches = [sig63, sig47]
    for match, expected in zip(sigs.signatures(), expected_matches):
        ss = load_one_signature_from_json(expected, ksize=31)
        assert match == ss


def test_prefetch_matches_to_dir(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with --save-matches to a directory
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")
    ss63 = load_one_signature_from_json(sig63)
    ss47 = load_one_signature_from_json(sig47)

    matches_out = c.output("matches_dir/")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "--save-matches",
        matches_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(matches_out)
    assert os.path.isdir(matches_out)

    sigs = sourmash.load_file_as_signatures(matches_out)

    match_sigs = list(sigs)
    assert ss63 in match_sigs
    assert ss47 in match_sigs
    assert len(match_sigs) == 2


def test_prefetch_matches_to_sig_gz(runtmp, linear_gather):
    c = runtmp

    import gzip

    # test a basic prefetch, with --save-matches to a sig.gz file
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")
    ss63 = load_one_signature_from_json(sig63)
    ss47 = load_one_signature_from_json(sig47)

    matches_out = c.output("matches.sig.gz")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "--save-matches",
        matches_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(matches_out)
    assert os.path.isfile(matches_out)

    with gzip.open(matches_out, "rt") as fp:
        # can we read this as a gz file?
        fp.read()

    sigs = sourmash.load_file_as_signatures(matches_out)

    match_sigs = list(sigs)
    assert ss63 in match_sigs
    assert ss47 in match_sigs
    assert len(match_sigs) == 2


def test_prefetch_matches_to_zip(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with --save-matches to a zipfile
    import zipfile

    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")
    ss63 = load_one_signature_from_json(sig63)
    ss47 = load_one_signature_from_json(sig47)

    matches_out = c.output("matches.zip")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "--save-matches",
        matches_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(matches_out)
    assert os.path.isfile(matches_out)

    with zipfile.ZipFile(matches_out, "r") as fp:
        # can we read this as a .zip file?
        for zi in fp.infolist():
            pass

    sigs = sourmash.load_file_as_signatures(matches_out)

    match_sigs = list(sigs)
    assert ss63 in match_sigs
    assert ss47 in match_sigs
    assert len(match_sigs) == 2


def test_prefetch_matching_hashes(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with --save-matches
    utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    matches_out = c.output("matches.sig")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        "--save-matching-hashes",
        matches_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(matches_out)

    ss47 = load_one_signature_from_json(sig47, ksize=31)
    ss63 = load_one_signature_from_json(sig63, ksize=31)
    matches = set(ss47.minhash.hashes) & set(ss63.minhash.hashes)

    intersect = ss47.minhash.copy_and_clear()
    intersect.add_many(matches)

    ss = load_one_signature_from_json(matches_out)
    assert ss.name.endswith("-known")
    assert ss.minhash == intersect


def test_prefetch_nomatch_hashes(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with --save-matches
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    nomatch_out = c.output("unmatched_hashes.sig")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        "--save-unmatched-hashes",
        nomatch_out,
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(nomatch_out)

    ss47 = load_one_signature_from_json(sig47, ksize=31)
    ss63 = load_one_signature_from_json(sig63, ksize=31)

    remain = ss47.minhash.to_mutable()
    remain.remove_many(ss63.minhash.hashes)

    ss = load_one_signature_from_json(nomatch_out)
    assert ss.name.endswith("-unknown")
    assert ss.minhash == remain


def test_prefetch_no_num_query(runtmp, linear_gather):
    c = runtmp

    # can't do prefetch with num signatures for query
    sig47 = utils.get_test_data("num/47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        c.run_sourmash("prefetch", "-k", "31", sig47, sig63, sig47, linear_gather)

    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status != 0


def test_prefetch_no_num_subj(runtmp, linear_gather):
    c = runtmp

    # can't do prefetch with num signatures for query; no matches!
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("num/63.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        c.run_sourmash("prefetch", "-k", "31", sig47, sig63, linear_gather)

    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status != 0
    assert (
        "ERROR in prefetch: after picklists and patterns, no signatures to search!?"
        in c.last_result.err
    )


def test_prefetch_db_fromfile(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    from_file = c.output("from-list.txt")

    with open(from_file, "w") as fp:
        print(sig63, file=fp)
        print(sig2, file=fp)
        print(sig47, file=fp)

    c.run_sourmash(
        "prefetch", "-k", "31", sig47, linear_gather, "--db-from-file", from_file
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0

    assert (
        "WARNING: no output(s) specified! Nothing will be saved from this prefetch!"
        in c.last_result.err
    )
    assert "selecting specified query k=31" in c.last_result.err
    assert (
        "loaded query: NC_009665.1 Shewanella baltica... (k=31, DNA)"
        in c.last_result.err
    )
    assert (
        "query sketch has scaled=1000; will be dynamically downsampled as needed"
        in c.last_result.err
    )

    assert "total of 2 matching signatures." in c.last_result.err
    assert (
        "of 5177 distinct query hashes, 5177 were found in matches above threshold."
        in c.last_result.err
    )
    assert "a total of 0 query hashes remain unmatched." in c.last_result.err


def test_prefetch_no_db(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch with no databases/signatures
    sig47 = utils.get_test_data("47.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        c.run_sourmash("prefetch", "-k", "31", sig47, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status != 0
    assert "ERROR: no databases or signatures to search!?" in c.last_result.err


def test_prefetch_check_scaled_bounds_negative(runtmp, linear_gather):
    c = runtmp

    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    with pytest.raises(SourmashCommandFailed) as exc:
        c.run_sourmash(
            "prefetch",
            "-k",
            "31",
            sig47,
            sig63,
            sig2,
            sig47,
            "--scaled",
            "-5",
            linear_gather,
        )

    assert "ERROR: scaled value must be positive" in str(exc.value)


def test_prefetch_check_scaled_bounds_less_than_minimum(runtmp, linear_gather):
    c = runtmp

    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    with pytest.raises(SourmashCommandFailed) as exc:
        c.run_sourmash(
            "prefetch",
            "-k",
            "31",
            sig47,
            sig63,
            sig2,
            sig47,
            "--scaled",
            "50",
            linear_gather,
        )

    assert "WARNING: scaled value should be >= 100. Continuing anyway." in str(
        exc.value
    )


def test_prefetch_check_scaled_bounds_more_than_maximum(runtmp, linear_gather):
    c = runtmp

    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    with pytest.raises(SourmashCommandFailed) as exc:
        c.run_sourmash(
            "prefetch",
            "-k",
            "31",
            sig47,
            sig63,
            sig2,
            sig47,
            "--scaled",
            "1e9",
            linear_gather,
        )

    assert "WARNING: scaled value should be <= 1e6. Continuing anyway." in str(
        exc.value
    )


def test_prefetch_downsample_scaled(runtmp, linear_gather):
    c = runtmp

    # test --scaled
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "--scaled",
        "1e5",
        linear_gather,
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert "downsampling query from scaled=1000 to 10000" in c.last_result.err


def test_prefetch_downsample_multiple(runtmp, linear_gather):
    # test multiple different downsamplings in prefetch code
    query_sig = utils.get_test_data("GCF_000006945.2-s500.sig")

    # load in the hashes and do split them into four bins, randomly.
    ss = load_one_signature_from_json(query_sig)
    hashes = list(ss.minhash.hashes)

    random.seed(a=1)  # fix seed so test is reproducible
    random.shuffle(hashes)

    # split into 4 bins:
    mh_bins = [ss.minhash.copy_and_clear() for i in range(4)]
    for i, hashval in enumerate(hashes):
        mh_bins[i % 4].add_hash(hashval)

    # downsample with different scaleds; initial scaled is 500, note.
    mh_bins[0] = mh_bins[0].downsample(scaled=750)
    mh_bins[1] = mh_bins[1].downsample(scaled=600)
    mh_bins[2] = mh_bins[2].downsample(scaled=1000)
    mh_bins[3] = mh_bins[3].downsample(scaled=650)

    gathersigs = []
    for i in range(4):
        binsig = SourmashSignature(mh_bins[i], name=f"bin{i}")

        with open(runtmp.output(f"bin{i}.sig"), "wb") as fp:
            save_signatures_to_json([binsig], fp)

        gathersigs.append(f"bin{i}.sig")

    runtmp.sourmash("prefetch", linear_gather, query_sig, *gathersigs)

    print(runtmp.last_result.out)
    print(runtmp.last_result.err)

    assert (
        "final scaled value (max across query and all matches) is 1000"
        in runtmp.last_result.err
    )


def test_prefetch_empty(runtmp, linear_gather):
    c = runtmp

    # test --scaled
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        c.run_sourmash(
            "prefetch",
            "-k",
            "31",
            sig47,
            sig63,
            sig2,
            sig47,
            "--scaled",
            "1e9",
            linear_gather,
        )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status != 0
    assert "no query hashes!? exiting." in c.last_result.err


def test_prefetch_basic_many_sigs(runtmp, linear_gather):
    c = runtmp

    # test what happens with many (and duplicate) signatures
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    manysigs = [sig63, sig2, sig47] * 5

    c.run_sourmash("prefetch", "-k", "31", sig47, *manysigs, linear_gather)
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert "total of 10 matching signatures so far." in c.last_result.err
    assert "total of 10 matching signatures." in c.last_result.err
    assert (
        "of 5177 distinct query hashes, 5177 were found in matches above threshold."
        in c.last_result.err
    )
    assert "a total of 0 query hashes remain unmatched." in c.last_result.err


def test_prefetch_with_picklist(runtmp):
    # test 'sourmash prefetch' with picklists
    gcf_sigs = glob.glob(utils.get_test_data("gather/GCF*.sig"))
    metag_sig = utils.get_test_data("gather/combined.sig")
    picklist = utils.get_test_data("gather/thermotoga-picklist.csv")

    runtmp.sourmash(
        "prefetch", metag_sig, *gcf_sigs, "--picklist", f"{picklist}:md5:md5"
    )

    err = runtmp.last_result.err
    print(err)
    assert "for given picklist, found 3 matches to 9 distinct values" in err
    # these are the different ksizes
    assert "WARNING: 6 missing picklist values." in err

    out = runtmp.last_result.out
    print(out)

    assert "total of 3 matching signatures." in err
    assert (
        "of 1466 distinct query hashes, 453 were found in matches above threshold."
        in err
    )
    assert "a total of 1013 query hashes remain unmatched." in err


def test_prefetch_with_picklist_exclude(runtmp):
    # test 'sourmash prefetch' with picklists, exclude
    gcf_sigs = glob.glob(utils.get_test_data("gather/GCF*.sig"))
    metag_sig = utils.get_test_data("gather/combined.sig")
    picklist = utils.get_test_data("gather/thermotoga-picklist.csv")

    runtmp.sourmash(
        "prefetch", metag_sig, *gcf_sigs, "--picklist", f"{picklist}:md5:md5:exclude"
    )

    err = runtmp.last_result.err
    print(err)
    assert "for given picklist, found 9 matches by excluding 9 distinct values" in err
    # these are the different ksizes

    out = runtmp.last_result.out
    print(out)

    assert "total of 9 matching signatures." in err
    assert (
        "of 1466 distinct query hashes, 1013 were found in matches above threshold."
        in err
    )
    assert "a total of 453 query hashes remain unmatched." in err


def test_prefetch_with_pattern_include(runtmp):
    # test 'sourmash prefetch' with --include-db-pattern
    gcf_sigs = glob.glob(utils.get_test_data("gather/GCF*.sig"))
    metag_sig = utils.get_test_data("gather/combined.sig")

    runtmp.sourmash("prefetch", metag_sig, *gcf_sigs, "--include", "thermotoga")

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "total of 3 matching signatures." in err
    assert (
        "of 1466 distinct query hashes, 453 were found in matches above threshold."
        in err
    )
    assert "a total of 1013 query hashes remain unmatched." in err


def test_prefetch_with_pattern_exclude(runtmp):
    # test 'sourmash prefetch' with --exclude-db-pattern
    gcf_sigs = glob.glob(utils.get_test_data("gather/GCF*.sig"))
    metag_sig = utils.get_test_data("gather/combined.sig")

    runtmp.sourmash("prefetch", metag_sig, *gcf_sigs, "--exclude", "thermotoga")

    err = runtmp.last_result.err
    print(err)

    out = runtmp.last_result.out
    print(out)

    assert "total of 9 matching signatures." in err
    assert (
        "of 1466 distinct query hashes, 1013 were found in matches above threshold."
        in err
    )
    assert "a total of 453 query hashes remain unmatched." in err


def test_prefetch_output_with_abundance(runtmp, prefetch_gather, linear_gather):
    c = runtmp
    query = utils.get_test_data("gather-abund/reads-s10x10-s11.sig")
    against = utils.get_test_data("gather-abund/genome-s10.fa.gz.sig")

    c.run_sourmash(
        "prefetch",
        linear_gather,
        query,
        against,
        "--save-matching-hashes",
        c.output("match-hash.sig"),
        "--save-unmatched-hashes",
        c.output("nomatch-hash.sig"),
    )

    print(c.last_result.out)

    assert os.path.exists(c.output("match-hash.sig"))
    ss = list(sourmash.load_file_as_signatures(c.output("match-hash.sig")))[0]
    assert ss.minhash.track_abundance

    assert os.path.exists(c.output("nomatch-hash.sig"))
    ss = list(sourmash.load_file_as_signatures(c.output("nomatch-hash.sig")))[0]
    assert ss.minhash.track_abundance


def test_prefetch_ani_csv_out(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with CSV output
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    csvout = c.output("out.csv")

    c.run_sourmash(
        "prefetch", "-k", "31", sig47, sig63, sig2, sig47, "-o", csvout, linear_gather
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(csvout)

    prefetch_result_names = PrefetchResult.prefetch_write_cols
    exp1 = {
        "q_ani": "0.9771552502238963",
        "m_ani": "0.9767860811200507",
        "ac_ani": "0.9769706656719734",
        "mc_ani": "0.9771552502238963",
        "pfn": "False",
    }
    exp2 = {
        "q_ani": "1.0",
        "m_ani": "1.0",
        "ac_ani": "1.0",
        "mc_ani": "1.0",
        "pfn": "False",
    }
    expected_ani_vals = [exp1, exp2]
    with open(csvout, newline="") as fp:
        r = csv.DictReader(fp)
        for row, expected in zip(r, expected_ani_vals):
            print(row)
            assert prefetch_result_names == list(row.keys())
            assert approx_eq(row["query_containment_ani"], expected["q_ani"])
            assert approx_eq(row["match_containment_ani"], expected["m_ani"])
            assert approx_eq(row["max_containment_ani"], expected["mc_ani"])
            assert approx_eq(row["average_containment_ani"], expected["ac_ani"])
            assert row["potential_false_negative"] == expected["pfn"]


def test_prefetch_ani_csv_out_estimate_ci(runtmp, linear_gather):
    c = runtmp

    # test a basic prefetch, with CSV output
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    csvout = c.output("out.csv")

    c.run_sourmash(
        "prefetch",
        "-k",
        "31",
        sig47,
        sig63,
        sig2,
        sig47,
        "-o",
        csvout,
        linear_gather,
        "--estimate-ani-ci",
    )
    print(c.last_result.status)
    print(c.last_result.out)
    print(c.last_result.err)

    assert c.last_result.status == 0
    assert os.path.exists(csvout)
    prefetch_result_names_ci = PrefetchResult.prefetch_write_cols_ci

    exp1 = {
        "q_ani": "0.9771552502238963",
        "m_ani": "0.9767860811200507",
        "q_ani_low": "0.9762537506990911",
        "q_ani_high": "0.9780336875157754",
        "m_ani_low": "0.9758801604653301",
        "m_ani_high": "0.9776692390768575",
        "ac_ani": "0.9769706656719734",
        "mc_ani": "0.9771552502238963",
        "pfn": "False",
    }
    exp2 = {
        "q_ani": "1.0",
        "m_ani": "1.0",
        "q_ani_low": "1.0",
        "q_ani_high": "1.0",
        "m_ani_low": "1.0",
        "m_ani_high": "1.0",
        "ac_ani": "1.0",
        "mc_ani": "1.0",
        "pfn": "False",
    }

    expected_ani_vals = [exp1, exp2]
    with open(csvout, newline="") as fp:
        r = csv.DictReader(fp)
        for row, expected in zip(r, expected_ani_vals):
            print(row)
            assert prefetch_result_names_ci == list(row.keys())
            assert approx_eq(row["query_containment_ani"], expected["q_ani"])
            assert approx_eq(row["query_containment_ani_low"], expected["q_ani_low"])
            assert approx_eq(row["query_containment_ani_high"], expected["q_ani_high"])
            assert approx_eq(row["match_containment_ani"], expected["m_ani"])
            assert approx_eq(row["match_containment_ani_low"], expected["m_ani_low"])
            assert approx_eq(row["match_containment_ani_high"], expected["m_ani_high"])
            assert approx_eq(row["max_containment_ani"], expected["mc_ani"])
            assert approx_eq(row["average_containment_ani"], expected["ac_ani"])
            assert row["potential_false_negative"] == expected["pfn"]


def test_prefetch_ani_containment_asymmetry(runtmp):
    # test contained_by asymmetries, viz #2215
    query_sig = utils.get_test_data("47.fa.sig")
    merged_sig = utils.get_test_data("47-63-merge.sig")

    runtmp.sourmash("prefetch", query_sig, merged_sig, "-o", "query-in-merged.csv")
    runtmp.sourmash("prefetch", merged_sig, query_sig, "-o", "merged-in-query.csv")

    with sourmash_args.FileInputCSV(runtmp.output("query-in-merged.csv")) as r:
        query_in_merged = list(r)[0]

    with sourmash_args.FileInputCSV(runtmp.output("merged-in-query.csv")) as r:
        merged_in_query = list(r)[0]

    assert query_in_merged["query_containment_ani"] == "1.0"
    assert query_in_merged["match_containment_ani"] == "0.9865155060423993"
    assert query_in_merged["average_containment_ani"] == "0.9932577530211997"

    assert merged_in_query["match_containment_ani"] == "1.0"
    assert merged_in_query["query_containment_ani"] == "0.9865155060423993"
    assert merged_in_query["average_containment_ani"] == "0.9932577530211997"



"""Define the command line interface for sourmash storage

The top level CLI is defined in ../__init__.py. This module defines the CLI for
`sourmash storage` operations.
"""

from . import convert
from ..utils import command_list
from argparse import SUPPRESS, RawDescriptionHelpFormatter
import os
import sys


def subparser(subparsers):
    subparser = subparsers.add_parser(
        "storage", formatter_class=RawDescriptionHelpFormatter, usage=SUPPRESS
    )
    desc = "Operations\n"
    clidir = os.path.dirname(__file__)
    ops = command_list(clidir)
    for subcmd in ops:
        docstring = getattr(sys.modules[__name__], subcmd).__doc__
        helpstring = f"sourmash storage {subcmd:s} --help"
        desc += f"        {helpstring:33s} {docstring:s}\n"
    s = subparser.add_subparsers(
        title="Storage utilities",
        dest="subcmd",
        metavar="subcmd",
        help=SUPPRESS,
        description=desc,
    )
    for subcmd in ops:
        getattr(sys.modules[__name__], subcmd).subparser(s)
    subparser._action_groups.reverse()
    subparser._optionals.title = "Options"



"Test SBT code."

import json
import shutil
import os

import pytest

import sourmash
from sourmash import load_one_signature, SourmashSignature, load_file_as_signatures
from sourmash.exceptions import IndexNotSupported
from sourmash.sbt import SBT, GraphFactory, Leaf, Node
from sourmash.sbtmh import SigLeaf, load_sbt_index
from sourmash.sbt_storage import FSStorage, RedisStorage, IPFSStorage, ZipStorage
from sourmash.search import make_jaccard_search_query
from sourmash.picklist import SignaturePicklist, PickStyle

import sourmash_tst_utils as utils


def test_simple(runtmp, n_children):
    factory = GraphFactory(5, 100, 3)
    root = SBT(factory, d=n_children)

    leaf1 = Leaf("a", factory())
    leaf1.data.count("AAAAA")
    leaf1.data.count("AAAAT")
    leaf1.data.count("AAAAC")

    leaf2 = Leaf("b", factory())
    leaf2.data.count("AAAAA")
    leaf2.data.count("AAAAT")
    leaf2.data.count("AAAAG")

    leaf3 = Leaf("c", factory())
    leaf3.data.count("AAAAA")
    leaf3.data.count("AAAAT")
    leaf3.data.count("CAAAA")

    leaf4 = Leaf("d", factory())
    leaf4.data.count("AAAAA")
    leaf4.data.count("CAAAA")
    leaf4.data.count("GAAAA")

    leaf5 = Leaf("e", factory())
    leaf5.data.count("AAAAA")
    leaf5.data.count("AAAAT")
    leaf5.data.count("GAAAA")

    root.add_node(leaf1)
    root.add_node(leaf2)
    root.add_node(leaf3)
    root.add_node(leaf4)
    root.add_node(leaf5)

    # return True if leaf node contains nodegraph w/kmer
    def search_kmer(leaf, kmer):
        return leaf.data.get(kmer)

    leaves = [leaf1, leaf2, leaf3, leaf4, leaf5]
    kmers = ["AAAAA", "AAAAT", "AAAAG", "CAAAA", "GAAAA"]

    # define an exhaustive search function that looks in all the leaf nodes.
    def search_kmer_in_list(kmer):
        x = []
        for l in leaves:
            if l.data.get(kmer):
                x.append(l)

        return set(x)

    # for all k-mers, ensure that tree._find_nodes matches the exhaustive
    # search.
    for kmer in kmers:
        assert set(root._find_nodes(search_kmer, kmer)) == search_kmer_in_list(kmer)

    print("-----")
    print([x.metadata for x in root._find_nodes(search_kmer, "AAAAA")])
    print([x.metadata for x in root._find_nodes(search_kmer, "AAAAT")])
    print([x.metadata for x in root._find_nodes(search_kmer, "AAAAG")])
    print([x.metadata for x in root._find_nodes(search_kmer, "CAAAA")])
    print([x.metadata for x in root._find_nodes(search_kmer, "GAAAA")])

    # save SBT to a directory and then reload
    root.save(runtmp.output("demo"))
    root = SBT.load(runtmp.output("demo"))

    for kmer in kmers:
        new_result = {str(r) for r in root._find_nodes(search_kmer, kmer)}
        print(*new_result, sep="\n")

        assert new_result == {str(r) for r in search_kmer_in_list(kmer)}


def test_longer_search(n_children):
    ksize = 5
    factory = GraphFactory(ksize, 100, 3)
    root = SBT(factory, d=n_children)

    leaf1 = Leaf("a", factory())
    leaf1.data.count("AAAAA")
    leaf1.data.count("AAAAT")
    leaf1.data.count("AAAAC")

    leaf2 = Leaf("b", factory())
    leaf2.data.count("AAAAA")
    leaf2.data.count("AAAAT")
    leaf2.data.count("AAAAG")

    leaf3 = Leaf("c", factory())
    leaf3.data.count("AAAAA")
    leaf3.data.count("AAAAT")
    leaf3.data.count("CAAAA")

    leaf4 = Leaf("d", factory())
    leaf4.data.count("AAAAA")
    leaf4.data.count("CAAAA")
    leaf4.data.count("GAAAA")

    leaf5 = Leaf("e", factory())
    leaf5.data.count("AAAAA")
    leaf5.data.count("AAAAT")
    leaf5.data.count("GAAAA")

    root.add_node(leaf1)
    root.add_node(leaf2)
    root.add_node(leaf3)
    root.add_node(leaf4)
    root.add_node(leaf5)

    def kmers(k, seq):
        for start in range(len(seq) - k + 1):
            yield seq[start : start + k]

    def search_transcript(node, seq, threshold):
        presence = [node.data.get(kmer) for kmer in kmers(ksize, seq)]
        if sum(presence) >= int(threshold * (len(seq) - ksize + 1)):
            return 1
        return 0

    try1 = [x.metadata for x in root._find_nodes(search_transcript, "AAAAT", 1.0)]
    assert set(try1) == set(["a", "b", "c", "e"]), try1  # no 'd'

    try2 = [x.metadata for x in root._find_nodes(search_transcript, "GAAAAAT", 0.6)]
    assert set(try2) == set(["a", "b", "c", "d", "e"])

    try3 = [x.metadata for x in root._find_nodes(search_transcript, "GAAAA", 1.0)]
    assert set(try3) == set(["d", "e"]), try3


# @pytest.mark.parametrize("old_version", ["v1", "v2", "v3", "v4", "v5"])
@pytest.mark.parametrize("old_version", ["v3", "v4", "v5"])
def test_tree_old_load(old_version):
    tree_old = SBT.load(
        utils.get_test_data(f"{old_version}.sbt.json"), leaf_loader=SigLeaf.load
    )

    tree_cur = SBT.load(utils.get_test_data("v6.sbt.json"), leaf_loader=SigLeaf.load)

    testdata1 = utils.get_test_data(utils.SIG_FILES[0])
    to_search = load_one_signature(testdata1)

    print(list(tree_old.leaves()))

    # note: earlier versions of this test did containment on
    # the num MinHash in `to_search`, which doesn't work properly.
    # (See test_sbt_no_containment_on_num for test). So, to
    # fix the test for the new search API, we had to adjust
    # the threshold.
    search_obj = make_jaccard_search_query(threshold=0.05)
    results_old = {str(s.signature) for s in tree_old.find(search_obj, to_search)}
    results_cur = {str(s.signature) for s in tree_cur.find(search_obj, to_search)}

    assert results_old == results_cur
    assert len(results_old) == 4


def test_load_future(tmpdir):
    with open(str(tmpdir.join("v9999.sbt.json")), "w") as f:
        json.dump({"version": 9999}, f)

    with pytest.raises(IndexNotSupported) as excinfo:
        SBT.load(str(tmpdir.join("v9999.sbt.json")))

    assert "index format is not supported" in str(excinfo.value)


def test_tree_save_load(runtmp, n_children):
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=n_children)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    tree.save(runtmp.output("demo"))
    tree = SBT.load(runtmp.output("demo"), leaf_loader=SigLeaf.load)

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    new_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*new_result, sep="\n")

    assert old_result == new_result


def test_search_minhashes():
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)

    to_search = next(iter(tree.leaves()))

    # this fails if 'search_obj' is calc containment and not similarity.
    search_obj = make_jaccard_search_query(threshold=0.08)
    results = tree.find(search_obj, to_search.data)

    n = 0
    for n, sr in enumerate(results):
        assert to_search.data.jaccard(sr.signature) >= 0.08

    assert n == 1


def test_binary_nary_tree():
    factory = GraphFactory(31, 1e5, 4)
    trees = {}
    trees[2] = SBT(factory)
    trees[5] = SBT(factory, d=5)
    trees[10] = SBT(factory, d=10)

    n_leaves = 0
    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        for tree in trees.values():
            tree.add_node(leaf)
        to_search = leaf
        n_leaves += 1

    assert all([len(list(t.leaves())) == n_leaves for t in trees.values()])

    results = {}
    print("*" * 60)
    print(f"{to_search.metadata}:")
    for d, tree in trees.items():
        search_obj = make_jaccard_search_query(threshold=0.1)
        results[d] = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*results[2], sep="\n")

    assert results[2] == results[5]
    assert results[5] == results[10]


def test_sbt_combine(n_children):
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=n_children)
    tree_1 = SBT(factory, d=n_children)
    tree_2 = SBT(factory, d=n_children)

    n_leaves = 0
    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        if n_leaves < 4:
            tree_1.add_node(leaf)
        else:
            tree_2.add_node(leaf)
        n_leaves += 1

    tree_1.combine(tree_2)

    t1_leaves = {str(l) for l in tree_1.leaves()}
    t_leaves = {str(l) for l in tree.leaves()}

    assert len(t1_leaves) == n_leaves
    assert len(t_leaves) == len(t1_leaves)
    assert t1_leaves == t_leaves

    to_search = load_one_signature(utils.get_test_data(utils.SIG_FILES[0]))
    search_obj = make_jaccard_search_query(threshold=0.1)
    t1_result = {str(s.signature) for s in tree_1.find(search_obj, to_search)}
    tree_result = {str(s.signature) for s in tree.find(search_obj, to_search)}
    assert t1_result == tree_result

    # TODO: save and load both trees

    # check if adding a new node will use the next empty position
    next_empty = 0
    for n, (d, _) in enumerate(tree_1):
        if n != d:
            next_empty = n
            break
    if not next_empty:
        next_empty = n + 1

    tree_1.add_node(SigLeaf(to_search.name, to_search))
    assert tree_1.next_node == next_empty


def test_sbt_fsstorage(runtmp):
    factory = GraphFactory(31, 1e5, 4)
    # with utils.TempDirectory() as location:
    tree = SBT(factory)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))

        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    with FSStorage(runtmp.location, ".fstree") as storage:
        tree.save(runtmp.output("tree.sbt.json"), storage=storage)

    tree = SBT.load(runtmp.output("tree.sbt.json"), leaf_loader=SigLeaf.load)
    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    new_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*new_result, sep="\n")

    assert old_result == new_result

    assert os.path.exists(runtmp.output(tree.storage.subdir))
    assert os.path.exists(runtmp.output(".fstree"))


def test_sbt_zipstorage(tmpdir):
    # create tree, save to a zip, then load and search.
    factory = GraphFactory(31, 1e5, 4)

    tree = SBT(factory)

    for f in utils.SIG_FILES:
        sig = next(load_file_as_signatures(utils.get_test_data(f)))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    with ZipStorage(str(tmpdir.join("tree.sbt.zip")), mode="w") as storage:
        tree.save(str(tmpdir.join("tree.sbt.json")), storage=storage)

    with ZipStorage(str(tmpdir.join("tree.sbt.zip"))) as storage:
        tree = SBT.load(
            str(tmpdir.join("tree.sbt.json")), leaf_loader=SigLeaf.load, storage=storage
        )

        print("*" * 60)
        print(f"{to_search.metadata}:")
        search_obj = make_jaccard_search_query(threshold=0.1)
        new_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
        print(*new_result, sep="\n")

        assert old_result == new_result


def test_sbt_ipfsstorage(runtmp):
    ipfshttpclient = pytest.importorskip("ipfshttpclient")

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))

        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    try:
        with IPFSStorage() as storage:
            tree.save(runtmp.output("tree.sbt.json"), storage=storage)
    except ipfshttpclient.exceptions.ConnectionError:
        pytest.xfail("ipfs not installed/functioning probably")

    with IPFSStorage() as storage:
        tree = SBT.load(
            runtmp.output("tree.sbt.json"), leaf_loader=SigLeaf.load, storage=storage
        )

        print("*" * 60)
        print(f"{to_search.metadata}:")
        search_obj = make_jaccard_search_query(threshold=0.1)
        new_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
        print(*new_result, sep="\n")

        assert old_result == new_result


def test_sbt_redisstorage(runtmp):
    redis = pytest.importorskip("redis")
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))

        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    try:
        with RedisStorage() as storage:
            tree.save(runtmp.output("tree.sbt.json"), storage=storage)
    except redis.exceptions.ConnectionError:
        pytest.xfail("Couldn't connect to redis server")

    with RedisStorage() as storage:
        tree = SBT.load(
            runtmp.output("tree.sbt.json"), leaf_loader=SigLeaf.load, storage=storage
        )

        print("*" * 60)
        print(f"{to_search.metadata}:")
        search_obj = make_jaccard_search_query(threshold=0.1)
        new_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
        print(*new_result, sep="\n")

        assert old_result == new_result


def test_save_zip(tmpdir):
    # load from zipped SBT, save to zipped SBT, and then search.
    testdata = utils.get_test_data("v6.sbt.zip")
    testsbt = tmpdir.join("v6.sbt.zip")
    newsbt = tmpdir.join("new.sbt.zip")

    shutil.copyfile(testdata, str(testsbt))

    tree = SBT.load(str(testsbt), leaf_loader=SigLeaf.load)
    tree.save(str(newsbt))
    assert newsbt.exists()

    new_tree = SBT.load(str(newsbt), leaf_loader=SigLeaf.load)
    assert isinstance(new_tree.storage, ZipStorage)
    assert new_tree.storage.list_sbts() == ["new.sbt.json"]

    to_search = load_one_signature(utils.get_test_data(utils.SIG_FILES[0]))

    print("*" * 60)
    print(f"{to_search}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search)}
    new_result = {str(s.signature) for s in new_tree.find(search_obj, to_search)}
    print(*new_result, sep="\n")

    assert old_result == new_result
    assert len(new_result) == 2


def test_load_zip(tmpdir):
    # search zipped SBT
    testdata = utils.get_test_data("v6.sbt.zip")
    testsbt = tmpdir.join("v6.sbt.zip")

    shutil.copyfile(testdata, str(testsbt))

    tree = SBT.load(str(testsbt), leaf_loader=SigLeaf.load)

    to_search = load_one_signature(utils.get_test_data(utils.SIG_FILES[0]))

    print("*" * 60)
    print(f"{to_search}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    new_result = {str(s.signature) for s in tree.find(search_obj, to_search)}
    print(*new_result, sep="\n")
    assert len(new_result) == 2


def test_load_zip_uncompressed(tmpdir):
    # uncompress zipped SBT into a tmpdir and search unpacked SBT
    import zipfile

    testdata = utils.get_test_data("v6.sbt.zip")
    testsbt = tmpdir.join("v6.sbt.json")

    with zipfile.ZipFile(testdata, "r") as z:
        z.extractall(str(tmpdir))

    tree = SBT.load(str(testsbt), leaf_loader=SigLeaf.load)

    to_search = load_one_signature(utils.get_test_data(utils.SIG_FILES[0]))

    print("*" * 60)
    print(f"{to_search}:")
    search_obj = make_jaccard_search_query(threshold=0.1)
    new_result = {str(s.signature) for s in tree.find(search_obj, to_search)}
    print(*new_result, sep="\n")
    assert len(new_result) == 2


def test_tree_repair():
    tree_repair = SBT.load(
        utils.get_test_data("leaves.sbt.json"), leaf_loader=SigLeaf.load
    )

    tree_cur = SBT.load(utils.get_test_data("v3.sbt.json"), leaf_loader=SigLeaf.load)

    testdata1 = utils.get_test_data(utils.SIG_FILES[0])
    to_search = load_one_signature(testdata1)

    search_obj = make_jaccard_search_query(threshold=0.1)
    results_repair = {str(s.signature) for s in tree_repair.find(search_obj, to_search)}
    results_cur = {str(s.signature) for s in tree_cur.find(search_obj, to_search)}

    assert results_repair == results_cur
    assert len(results_repair) == 2


def test_tree_repair_insert():
    tree_repair = SBT.load(
        utils.get_test_data("leaves.sbt.json"), leaf_loader=SigLeaf.load
    )

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree_repair.add_node(leaf)

    for pos, node in tree_repair:
        # Every parent of a node must be an internal node (and not a leaf),
        # except for node 0 (the root), whose parent is None.
        if pos != 0:
            assert isinstance(tree_repair.parent(pos).node, Node)

        # Leaf nodes can't have children
        if isinstance(node, Leaf):
            assert all(c.node is None for c in tree_repair.children(pos))


def test_save_sparseness(runtmp, n_children):
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=n_children)

    for f in utils.SIG_FILES:
        sig = load_one_signature(utils.get_test_data(f))
        leaf = SigLeaf(os.path.basename(f), sig)
        tree.add_node(leaf)
        to_search = leaf

    print("*" * 60)
    print(f"{to_search.metadata}:")

    search_obj = make_jaccard_search_query(threshold=0.1)
    old_result = {str(s.signature) for s in tree.find(search_obj, to_search.data)}
    print(*old_result, sep="\n")

    tree.save(runtmp.output("demo"), sparseness=1.0)
    tree_loaded = SBT.load(runtmp.output("demo"), leaf_loader=SigLeaf.load)
    assert all(not isinstance(n, Node) for _, n in tree_loaded)

    print("*" * 60)
    print(f"{to_search.metadata}:")
    new_result = {
        str(s.signature) for s in tree_loaded.find(search_obj, to_search.data)
    }
    print(*new_result, sep="\n")

    assert old_result == new_result

    for pos, node in tree_loaded:
        # Every parent of a node must be an internal node (and not a leaf),
        # except for node 0 (the root), whose parent is None.
        if pos != 0:
            assert isinstance(tree_loaded.parent(pos).node, Node)

        # Leaf nodes can't have children
        if isinstance(node, Leaf):
            assert all(c.node is None for c in tree_loaded.children(pos))


def test_sbt_as_index_select():
    # test 'select' method from Index base class.
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    xx = tree.select(ksize=31)
    assert xx == tree

    xx = tree.select(moltype="DNA")
    assert xx == tree

    xx = tree.select(abund=False)
    assert xx == tree

    with pytest.raises(ValueError):
        tree.select(ksize=21)

    with pytest.raises(ValueError):
        tree.select(moltype="protein")

    with pytest.raises(ValueError):
        tree.select(abund=True)


def test_sbt_as_index_select_picklist():
    # test 'select' method from Index base class with a picklist

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8")
    picklist.init(["09a08691"])

    # select on picklist
    tree = tree.select(picklist=picklist)
    siglist = list(tree.signatures())
    assert len(siglist) == 1

    ss = siglist[0]
    assert ss.minhash.ksize == 31
    assert ss.md5sum().startswith("09a08691c")


def test_sbt_as_index_select_picklist_exclude():
    # test 'select' method from Index base class with a picklist, exclude

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8", pickstyle=PickStyle.EXCLUDE)
    picklist.init(["09a08691"])

    # select on picklist
    tree = tree.select(picklist=picklist)
    siglist = list(tree.signatures())
    assert len(siglist) == 1

    ss = siglist[0]
    assert ss.minhash.ksize == 31
    assert ss.md5sum().startswith("38729c637")


def test_sbt_as_index_find_picklist():
    # test 'select' method from Index base class with a picklist

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8")
    picklist.init(["09a08691"])

    # run a 'find' with sig63, should find 47 and 63 both.
    search_obj = make_jaccard_search_query(do_containment=True, threshold=0.0)
    results = list(tree.find(search_obj, sig63))
    print(results)
    assert len(results) == 2

    # now, select on picklist and do another find...
    tree = tree.select(picklist=picklist)
    results = list(tree.find(search_obj, sig63))
    print(results)
    assert len(results) == 1

    # and check that it is the expected one!
    ss = results[0].signature
    assert ss.minhash.ksize == 31
    assert ss.md5sum().startswith("09a08691c")


def test_sbt_as_index_find_picklist_exclude():
    # test 'select' method from Index base class with a picklist

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8", pickstyle=PickStyle.EXCLUDE)
    picklist.init(["09a08691"])

    # run a 'find' with sig63, should find 47 and 63 both.
    search_obj = make_jaccard_search_query(do_containment=True, threshold=0.0)
    results = list(tree.find(search_obj, sig63))
    print(results)
    assert len(results) == 2

    # now, select on picklist and do another find...
    tree = tree.select(picklist=picklist)
    results = list(tree.find(search_obj, sig63))
    print(results)
    assert len(results) == 1

    # and check that it is the expected one!
    ss = results[0].signature
    assert ss.minhash.ksize == 31
    assert ss.md5sum().startswith("38729c637")


def test_sbt_as_index_find_picklist_twice():
    # test 'select' method from Index base class with a picklist

    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8")
    picklist.init(["09a08691"])

    # run a 'find' with sig63, should find 47 and 63 both.
    search_obj = make_jaccard_search_query(do_containment=True, threshold=0.0)
    results = list(tree.find(search_obj, sig63))
    print(results)
    assert len(results) == 2

    # now, select twice on picklists...
    tree = tree.select(picklist=picklist)

    with pytest.raises(ValueError):
        tree = tree.select(picklist=picklist)
        assert "we do not (yet) support multiple picklists for SBT databases" in str(
            exc
        )


def test_sbt_as_index_signatures():
    # test 'signatures' method from Index base class.
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"))
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"))

    tree.insert(sig47)
    tree.insert(sig63)

    xx = list(tree.signatures())
    assert len(xx) == 2

    assert sig47 in xx
    assert sig63 in xx


def test_sbt_gather_threshold_1():
    # test gather() method, in some detail
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig2 = load_one_signature(utils.get_test_data("2.fa.sig"), ksize=31)
    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"), ksize=31)
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"), ksize=31)

    tree.insert(sig47)
    tree.insert(sig63)
    tree.insert(sig2)

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(sig2.minhash.hashes.keys()))
    new_mh = sig2.minhash.copy_and_clear()

    # query with empty hashes
    assert not new_mh
    with pytest.raises(ValueError):
        tree.best_containment(SourmashSignature(new_mh))

    # add one hash
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 1

    result = tree.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig == sig2
    assert name is None

    # check with a threshold -> should be no results.
    with pytest.raises(ValueError):
        tree.best_containment(SourmashSignature(new_mh), threshold_bp=5000)

    # add three more hashes => length of 4
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 4

    result = tree.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig == sig2
    assert name is None

    # check with a too-high threshold -> should be no results.
    print("len mh", len(new_mh))
    with pytest.raises(ValueError):
        tree.best_containment(SourmashSignature(new_mh), threshold_bp=5000)


def test_sbt_gather_threshold_5():
    # test gather() method above threshold
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    sig2 = load_one_signature(utils.get_test_data("2.fa.sig"), ksize=31)
    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"), ksize=31)
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"), ksize=31)

    tree.insert(sig47)
    tree.insert(sig63)
    tree.insert(sig2)

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(sig2.minhash.hashes.keys()))
    new_mh = sig2.minhash.copy_and_clear()

    # add five hashes
    for i in range(5):
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())

    # should get a result with no threshold (any match at all is returned)
    result = tree.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig == sig2
    assert name is None

    # now, check with a threshold_bp that should be meet-able.
    tree.best_containment(SourmashSignature(new_mh), threshold_bp=5000)
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig == sig2
    assert name is None


@utils.in_tempdir
def test_gather_single_return(c):
    # test gather() number of returns
    sig2file = utils.get_test_data("2.fa.sig")
    sig47file = utils.get_test_data("47.fa.sig")
    sig63file = utils.get_test_data("63.fa.sig")

    sig2 = load_one_signature(sig2file, ksize=31)
    sig47 = load_one_signature(sig47file, ksize=31)
    sig63 = load_one_signature(sig63file, ksize=31)

    # construct SBT Database
    factory = GraphFactory(31, 1e5, 4)
    tree = SBT(factory, d=2)

    tree.insert(sig2)
    tree.insert(sig47)
    tree.insert(sig63)

    # now, run gather. how many results do we get, and are they in the
    # right order?
    result = tree.best_containment(sig63)
    print(result)
    assert result
    assert result.score == 1.0


def test_sbt_jaccard_ordering(runtmp):
    # this tests a tricky situation where for three sketches A, B, C,
    # |A intersect B| is greater than |A intersect C|
    # _but_
    # |A jaccard B| is less than |A intersect B|
    a = sourmash.MinHash(ksize=31, n=0, scaled=2)
    b = a.copy_and_clear()
    c = a.copy_and_clear()

    a.add_many([1, 2, 3, 4])
    b.add_many([1, 2, 3] + list(range(10, 30)))
    c.add_many([1, 5])

    def _intersect(x, y):
        return x.intersection_and_union_size(y)[0]

    print("a intersect b:", _intersect(a, b))
    print("a intersect c:", _intersect(a, c))
    print("a jaccard b:", a.jaccard(b))
    print("a jaccard c:", a.jaccard(c))
    assert _intersect(a, b) > _intersect(a, c)
    assert a.jaccard(b) < a.jaccard(c)

    # thresholds to use:
    assert a.jaccard(b) < 0.15
    assert a.jaccard(c) > 0.15

    # now - make signatures, try out :)
    ss_a = sourmash.SourmashSignature(a, name="A")
    ss_b = sourmash.SourmashSignature(b, name="B")
    ss_c = sourmash.SourmashSignature(c, name="C")

    factory = GraphFactory(31, 1e5, 4)
    db = SBT(factory, d=2)
    db.insert(ss_a)
    db.insert(ss_b)
    db.insert(ss_c)

    sr = db.search(ss_a, threshold=0.15)
    print(sr)
    assert len(sr) == 2
    assert sr[0].signature == ss_a
    assert sr[0].score == 1.0
    assert sr[1].signature == ss_c
    assert sr[1].score == 0.2


def test_sbt_protein_command_index(runtmp):
    c = runtmp

    # test command-line creation of SBT database with protein sigs
    sigfile1 = utils.get_test_data(
        "prot/protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    sigfile2 = utils.get_test_data(
        "prot/protein/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig"
    )

    db_out = c.output("protein.sbt.zip")

    c.run_sourmash(
        "index", db_out, sigfile1, sigfile2, "--scaled", "100", "-k", "19", "--protein"
    )

    # check to make sure .sbt.protein directory doesn't get created
    assert not os.path.exists(c.output(".sbt.protein"))

    db2 = load_sbt_index(db_out)

    sig1 = sourmash.load_one_signature(sigfile1)
    sig2 = sourmash.load_one_signature(sigfile2)

    # check reconstruction --
    mh_list = [x.minhash for x in db2.signatures()]
    assert len(mh_list) == 2
    assert sig1.minhash in mh_list
    assert sig2.minhash in mh_list

    # and search, gather
    results = db2.search(
        sig1,
        threshold=0.0,
        ignore_abundance=True,
        do_containment=False,
        best_only=False,
    )
    assert len(results) == 2

    result = db2.best_containment(sig2)
    assert result.score == 1.0
    assert result.location == db2._location
    assert result.location == db_out


@utils.in_tempdir
def test_sbt_protein_search_no_threshold(c):
    # test the '.search' method on SBTs w/no threshold
    sigfile1 = utils.get_test_data(
        "prot/protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    sigfile2 = utils.get_test_data(
        "prot/protein/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig"
    )

    db_out = c.output("protein.sbt.zip")

    c.run_sourmash(
        "index", db_out, sigfile1, sigfile2, "--scaled", "100", "-k", "19", "--protein"
    )

    db2 = load_sbt_index(db_out)

    sig1 = sourmash.load_one_signature(sigfile1)

    # and search, gather
    with pytest.raises(TypeError) as exc:
        db2.search(sig1)
    assert "'search' requires 'threshold'" in str(exc)


@utils.in_thisdir
def test_sbt_protein_command_search(c):
    # test command-line search/gather of SBT database with protein sigs
    sigfile1 = utils.get_test_data(
        "prot/protein/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    db_out = utils.get_test_data("prot/protein.sbt.zip")

    c.run_sourmash("search", sigfile1, db_out, "--threshold", "0.0")
    assert "2 matches" in c.last_result.out

    c.run_sourmash("gather", sigfile1, db_out)
    assert "found 1 matches total" in c.last_result.out
    assert "the recovered matches hit 100.0% of the query" in c.last_result.out


@utils.in_tempdir
def test_sbt_hp_command_index(c):
    # test command-line creation of SBT database with hp sigs
    sigfile1 = utils.get_test_data(
        "prot/hp/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    sigfile2 = utils.get_test_data(
        "prot/hp/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig"
    )

    db_out = c.output("hp.sbt.zip")

    c.run_sourmash(
        "index", db_out, sigfile1, sigfile2, "--scaled", "100", "-k", "19", "--hp"
    )

    db2 = load_sbt_index(db_out)

    sig1 = sourmash.load_one_signature(sigfile1)
    sig2 = sourmash.load_one_signature(sigfile2)

    # check reconstruction --
    mh_list = [x.minhash for x in db2.signatures()]
    assert len(mh_list) == 2
    assert sig1.minhash in mh_list
    assert sig2.minhash in mh_list

    # and search, gather
    results = db2.search(
        sig1,
        threshold=0.0,
        ignore_abundance=True,
        do_containment=False,
        best_only=False,
    )
    assert results

    result = db2.best_containment(sig2)
    assert result.score == 1.0
    assert result.location == db2._location
    assert result.location == db_out


@utils.in_thisdir
def test_sbt_hp_command_search(c):
    # test command-line search/gather of SBT database with hp sigs
    sigfile1 = utils.get_test_data(
        "prot/hp/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    db_out = utils.get_test_data("prot/hp.sbt.zip")

    c.run_sourmash("search", sigfile1, db_out, "--threshold", "0.0")
    assert "2 matches" in c.last_result.out

    c.run_sourmash("gather", sigfile1, db_out, "--threshold", "0.0")
    assert "found 1 matches total" in c.last_result.out
    assert "the recovered matches hit 100.0% of the query" in c.last_result.out


@utils.in_tempdir
def test_sbt_dayhoff_command_index(c):
    # test command-line creation of SBT database with dayhoff sigs
    sigfile1 = utils.get_test_data(
        "prot/dayhoff/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    sigfile2 = utils.get_test_data(
        "prot/dayhoff/GCA_001593935.1_ASM159393v1_protein.faa.gz.sig"
    )

    db_out = c.output("dayhoff.sbt.zip")

    c.run_sourmash(
        "index", db_out, sigfile1, sigfile2, "--scaled", "100", "-k", "19", "--dayhoff"
    )

    db2 = load_sbt_index(db_out)

    sig1 = sourmash.load_one_signature(sigfile1)
    sig2 = sourmash.load_one_signature(sigfile2)

    # check reconstruction --
    mh_list = [x.minhash for x in db2.signatures()]
    assert len(mh_list) == 2
    assert sig1.minhash in mh_list
    assert sig2.minhash in mh_list

    # and search, gather
    results = db2.search(
        sig1,
        threshold=0.0,
        ignore_abundance=True,
        do_containment=False,
        best_only=False,
    )
    assert len(results) == 2

    result = db2.best_containment(sig2)
    assert result.score == 1.0
    assert result.location == db2._location
    assert result.location == db_out


@utils.in_thisdir
def test_sbt_dayhoff_command_search(c):
    # test command-line search/gather of SBT database with dayhoff sigs
    sigfile1 = utils.get_test_data(
        "prot/dayhoff/GCA_001593925.1_ASM159392v1_protein.faa.gz.sig"
    )
    db_out = utils.get_test_data("prot/dayhoff.sbt.zip")

    c.run_sourmash("search", sigfile1, db_out, "--threshold", "0.0")
    assert "2 matches" in c.last_result.out

    c.run_sourmash("gather", sigfile1, db_out, "--threshold", "0.0")
    assert "found 1 matches total" in c.last_result.out
    assert "the recovered matches hit 100.0% of the query" in c.last_result.out


def test_sbt_node_cache():
    tree = SBT.load(
        utils.get_test_data("v6.sbt.json"), leaf_loader=SigLeaf.load, cache_size=1
    )

    testdata1 = utils.get_test_data(utils.SIG_FILES[0])
    to_search = load_one_signature(testdata1)

    # note: earlier versions of this test did containment on
    # the num MinHash in `to_search`, which doesn't work properly.
    # (See test_sbt_no_containment_on_num for test). So, to
    # fix the test for the new search API, we had to adjust
    # the threshold.
    search_obj = make_jaccard_search_query(threshold=0.05)
    results = list(tree.find(search_obj, to_search))
    assert len(results) == 4

    assert tree._nodescache.currsize == 1
    assert tree._nodescache.currsize == 1


def test_sbt_no_containment_on_num():
    tree = SBT.load(
        utils.get_test_data("v6.sbt.json"), leaf_loader=SigLeaf.load, cache_size=1
    )

    testdata1 = utils.get_test_data(utils.SIG_FILES[0])
    to_search = load_one_signature(testdata1)

    search_obj = make_jaccard_search_query(do_containment=True, threshold=0.05)
    with pytest.raises(TypeError) as exc:
        list(tree.find(search_obj, to_search))

    assert "this search requires a scaled signature" in str(exc)


def test_build_sbt_zip_with_dups(runtmp):
    dups_data = utils.get_test_data("duplicate-sigs")

    all_sigs = set(sourmash.load_file_as_signatures(dups_data))
    assert len(all_sigs) == 4

    runtmp.run_sourmash("index", "dups.sbt.zip", dups_data)
    outfile = runtmp.output("dups.sbt.zip")

    sbt_sigs = set(sourmash.load_file_as_signatures(outfile))
    assert len(sbt_sigs) == 4

    assert all_sigs == sbt_sigs


def test_build_sbt_zip_with_dups_exists(runtmp):
    dups_data = utils.get_test_data("duplicate-sigs")

    all_sigs = set(sourmash.load_file_as_signatures(dups_data))
    assert len(all_sigs) == 4

    runtmp.run_sourmash("index", "dups.sbt.zip", dups_data)
    outfile = runtmp.output("dups.sbt.zip")

    # run again, to see what happens :)
    runtmp.run_sourmash("index", "dups.sbt.zip", dups_data)
    outfile = runtmp.output("dups.sbt.zip")

    sbt_sigs = set(sourmash.load_file_as_signatures(outfile))
    assert len(sbt_sigs) == 4

    assert all_sigs == sbt_sigs


def test_build_sbt_json_with_dups(runtmp):
    dups_data = utils.get_test_data("duplicate-sigs")

    all_sigs = set(sourmash.load_file_as_signatures(dups_data))
    assert len(all_sigs) == 4

    runtmp.run_sourmash("index", "dups.sbt.json", dups_data)
    outfile = runtmp.output("dups.sbt.json")

    sbt_sigs = set(sourmash.load_file_as_signatures(outfile))
    assert len(sbt_sigs) == 4

    assert all_sigs == sbt_sigs


def test_build_sbt_json_with_dups_exists(runtmp):
    dups_data = utils.get_test_data("duplicate-sigs")

    all_sigs = set(sourmash.load_file_as_signatures(dups_data))
    assert len(all_sigs) == 4

    runtmp.run_sourmash("index", "dups.sbt.json", dups_data)
    outfile = runtmp.output("dups.sbt.json")

    # run again, see what happens!
    runtmp.run_sourmash("index", "dups.sbt.json", dups_data)
    outfile = runtmp.output("dups.sbt.json")

    sbt_sigs = set(sourmash.load_file_as_signatures(outfile))
    assert len(sbt_sigs) == 4

    assert all_sigs == sbt_sigs


def test_load_fail_on_file_not_dir(runtmp):
    # make sure the load function raises a ValueError for {filename}/sbt,
    # rather than a NotADirectoryError

    filename = runtmp.output("foo")
    with open(filename, "w") as fp:
        fp.write("something")

    with pytest.raises(ValueError):
        SBT.load(runtmp.output("foo/bar.sbt.json"))
